#include "OgrePathMapRenderTechnique.h"
#include "OgreTechniqueGroup.h"
#include "OgreIlluminationManager.h"
#include "OgrePMWeightComputeRenderingRun.h"



OgrePathMapRenderTechnique::OgrePathMapRenderTechnique(											
												SceneBlendFactor passBlendingSRC,
												SceneBlendFactor passBlendingDEST,
												Pass* pass,
												OgreRenderable* parentRenderable,
												OgreTechniqueGroup* parentTechniqueGroup)
							:RenderTechnique( parentRenderable, parentTechniqueGroup),
							OgreRenderTechnique(pass, parentRenderable, parentTechniqueGroup)
{	
	this->passBlendingSRC = passBlendingSRC;
	this->passBlendingDEST = passBlendingDEST;

	this->clusters = OgreIlluminationManager::getSingleton().getPathMapClusters(parentRenderable->getName());
	
	//insert new pass
	Ogre::Technique* techn = pass->getParent();
	Technique::PassIterator it = techn->getPassIterator();
	
	int index = 0;
	while(it.hasMoreElements())
	{
		if( it.getNext() == pass)
			break;
		index++;
		it.moveNext();
	}
	
	Pass* newpass = pathMapPass = techn->createPass();
	
	newpass->setVertexProgram("GTP/PathMap_VS");
	newpass->setFragmentProgram("GTP/PathMap_PS");
	
	//bind vertex program parameters
	GpuProgramParameters* Vparams = newpass->getVertexProgramParameters().getPointer();
	Vparams->setNamedAutoConstant("WorldViewProj", GpuProgramParameters::ACT_WORLDVIEWPROJ_MATRIX);	
	//bind fragment program parameters
	GpuProgramParameters* Fparams = newpass->getFragmentProgramParameters().getPointer();
	int prmxres = clusters->pathMapResolution * clusters->count;
	int prmyres = clusters->pathMapResolution;
	while(prmxres > 4096)
	{
		prmxres /= 2;
		prmyres *= 2;		
	}
	int prmnt[2] = {prmxres / clusters->pathMapResolution, prmyres / clusters->pathMapResolution};

	float halfPixel[2] = {0.5 / prmxres, 0.5 / prmyres};
	Vector4 pathMapParameters(prmnt[0],prmnt[1],halfPixel[0],halfPixel[1]);
	Fparams->setNamedConstant("prmAtlasTilesHalfPixel",pathMapParameters);
    
	unsigned int clustercount = OgreIlluminationManager::getSingleton().getPathMapClusterLengthsSize();
	Fparams->setNamedConstant("allClusterCount", (float) clustercount);
	PathMapClusters* clusters = OgreIlluminationManager::getSingleton().getPathMapClusters(parentOgreRenderable->getName());
	Fparams->setNamedConstant("clusterCount", (float) clusters->count);
	
	TextureUnitState* st = newpass->createTextureUnitState();		
	st->setTextureFiltering(TFO_BILINEAR);
	st->setTextureAddressingMode(TextureUnitState::TAM_CLAMP);
	st->setTextureBorderColour(ColourValue::Green);
	st->setTextureName(clusters->pathMapTextureFilename);

	createWeightIndexTexture();
	
	st = newpass->createTextureUnitState();		
	st->setTextureFiltering(TFO_NONE);
	st->setTextureAddressingMode(TextureUnitState::TAM_CLAMP);
	st->setTextureBorderColour(ColourValue::Red);
	st->setTextureName(weightIndexTexture->getName());

	st = newpass->createTextureUnitState();		
	st->setTextureFiltering(TFO_NONE);
	st->setTextureAddressingMode(TextureUnitState::TAM_CLAMP);
	st->setTextureBorderColour(ColourValue::Blue);

	st = newpass->createTextureUnitState();		
	st->setTextureFiltering(TFO_BILINEAR);
	st->setTextureAddressingMode(TextureUnitState::TAM_WRAP);
	st->setTextureBorderColour(ColourValue::Blue);
	st->setTextureName(pass->getTextureUnitState(0)->getTextureName());
	
	newpass->setSceneBlending(passBlendingSRC, passBlendingDEST);
	newpass->setDepthBias(1);
}

OgrePathMapRenderTechnique::~OgrePathMapRenderTechnique()
{

}

void OgrePathMapRenderTechnique::createWeightIndexTexture()
{
	PathMapClusters* clusters = OgreIlluminationManager::getSingleton().getPathMapClusters(parentOgreRenderable->getName());
	int width = clusters->count;
	TexturePtr texPtr = Ogre::TextureManager::getSingleton().createManual(this->parentOgreRenderable->getName() + "_PMWeightIndexTexture", 
																		"default",
																		TEX_TYPE_2D,
																		width,
																		1,
																		0,
																		0,
																		PF_FLOAT32_R,
																		TU_DYNAMIC_WRITE_ONLY);
	weightIndexTexture = texPtr.getPointer();

	float *weightIndices = new float[width];
	PixelBox lockBox(width, 1, 1, PF_FLOAT32_R, weightIndices);
	for(int j = 0; j< clusters->count; j++)
			weightIndices[j] = clusters->clusters[j];
	weightIndexTexture->getBuffer()->blitFromMemory(lockBox);
	delete[] weightIndices;
}

void OgrePathMapRenderTechnique::update(unsigned long frameNum)
{	
	LightList lights;
	SceneManager* sm = Root::getSingleton()._getCurrentSceneManager();
	sm->_populateLightList(OgreIlluminationManager::getSingleton().getMainCamera()->getPosition(), 1000, lights);

	for(int i = 0 ; i < 5 && i < lights.size(); i++)
	{
		String lightName = lights.at(i)->getName();
		OgreIlluminationManager::getSingleton().createPerLightRun(lightName, ILLUMRUN_PM_WEIGHTMAP);
		OgreIlluminationManager::getSingleton().updatePerLightRun(lightName, ILLUMRUN_PM_WEIGHTMAP, frameNum);		
	}
	OgrePMWeightComputeRenderingRun::sumWeights(frameNum);
	TextureUnitState* st = pathMapPass->getTextureUnitState(2);
	st->setTextureName(OgrePMWeightComputeRenderingRun::getPMWeightTextureName());
}



namespace PathMapParsers
{
	SceneBlendFactor convertBlendFactor(const String& param)
    {
        if (param == "one")
            return SBF_ONE;
        else if (param == "zero")
            return SBF_ZERO;
        else if (param == "dest_colour")
            return SBF_DEST_COLOUR;
        else if (param == "src_colour")
            return SBF_SOURCE_COLOUR;
        else if (param == "one_minus_dest_colour")
            return SBF_ONE_MINUS_DEST_COLOUR;
        else if (param == "one_minus_src_colour")
            return SBF_ONE_MINUS_SOURCE_COLOUR;
        else if (param == "dest_alpha")
            return SBF_DEST_ALPHA;
        else if (param == "src_alpha")
            return SBF_SOURCE_ALPHA;
        else if (param == "one_minus_dest_alpha")
            return SBF_ONE_MINUS_DEST_ALPHA;
        else if (param == "one_minus_src_alpha")
            return SBF_ONE_MINUS_SOURCE_ALPHA;       
    }

	void parsePassBlending(String& params, RenderTechniqueFactory* factory)
	{
		OgrePathMapRenderTechniqueFactory* f = (OgrePathMapRenderTechniqueFactory*) factory;
		StringVector vecparams = StringUtil::split(params, " \t");
		if(vecparams.size() == 1)
		{
			if (vecparams[0] == "none")
			{
                f->passBlendingSRC =  SBF_ONE;
				f->passBlendingDEST = SBF_ZERO;
			}
			if (vecparams[0] == "add")
			{
                f->passBlendingSRC =  SBF_ONE;
				f->passBlendingDEST = SBF_ONE;
			}
            else if (vecparams[0] == "modulate")
            {
				f->passBlendingSRC =  SBF_DEST_COLOUR;
				f->passBlendingDEST = SBF_ZERO;
			}
			else if (vecparams[0] == "colour_blend")
			{
				f->passBlendingSRC =  SBF_SOURCE_COLOUR;
				f->passBlendingDEST = SBF_ONE_MINUS_SOURCE_COLOUR;
			}
            else if (vecparams[0] == "alpha_blend")
            {
                f->passBlendingSRC =  SBF_SOURCE_ALPHA;
				f->passBlendingDEST = SBF_ONE_MINUS_SOURCE_ALPHA;
			}            
		}
		else if (vecparams.size() == 2)
		{
		  f->passBlendingSRC =  convertBlendFactor(vecparams[0]);
		  f->passBlendingDEST = convertBlendFactor(vecparams[1]) ;
		}		
	}	
}

OgrePathMapRenderTechniqueFactory::OgrePathMapRenderTechniqueFactory()
{
	typeName = "PathMap";
	using namespace PathMapParsers;
	this->attributeParsers.insert(AttribParserList::value_type("pass_blending", (ILLUM_ATTRIBUTE_PARSER) parsePassBlending));
}

OgreRenderTechnique* OgrePathMapRenderTechniqueFactory::createInstance( 
										IllumTechniqueParams* params,
										Pass* pass,
										OgreRenderable* parentRenderable,
										OgreTechniqueGroup* parentTechniqueGroup)
{	
	passBlendingSRC = SBF_ONE;
	passBlendingDEST = SBF_ONE;

	parseParams(params);

	OgrePathMapRenderTechnique* result = new OgrePathMapRenderTechnique(
												passBlendingSRC,
												passBlendingDEST,
												pass,
												parentRenderable,
												parentTechniqueGroup);	

	return result;
}

